from unittest.mock import patch

import pytest

from .corellium.exceptions import CorelliumApiException
from .corellium.types import Device, Instance, Project
from .driver import Corellium, CorelliumConsole, CorelliumPower
from jumpstarter.common import exceptions as jmp_exceptions


def test_driver_corellium_init_ok(monkeypatch):
    monkeypatch.setenv('CORELLIUM_API_HOST', 'api-host')
    monkeypatch.setenv('CORELLIUM_API_TOKEN', 'api-token')

    c = Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')

    assert '1' == c.project_id
    assert 'jmp'  == c.device_name
    assert 'kronos' == c.device_flavor
    assert '1.0' == c.device_os
    assert 'api-host'  == c.api.host
    assert 'api-token' == c.api.token


@pytest.mark.parametrize(
        'env,err',
        [
            (
                {},
                jmp_exceptions.ConfigurationError('Missing "CORELLIUM_API_HOST" environment variable')
            ),
            (
                {'CORELLIUM_API_HOST': '  '},
                jmp_exceptions.ConfigurationError('"CORELLIUM_API_HOST" environment variable is empty')
            ),
            (
                {'CORELLIUM_API_HOST': 'api-host'},
                jmp_exceptions.ConfigurationError('Missing "CORELLIUM_API_TOKEN" environment variable')
            ),
            (
                {'CORELLIUM_API_HOST': 'api-host', 'CORELLIUM_API_TOKEN': '   '},
                jmp_exceptions.ConfigurationError('"CORELLIUM_API_TOKEN" environment variable is empty')
            ),
        ])
def test_driver_corellium_init_error(monkeypatch, env, err):
    monkeypatch.delenv('CORELLIUM_API_HOST', raising=False)
    monkeypatch.delenv('CORELLIUM_API_TOKEN', raising=False)

    for k, v in env.items():
        monkeypatch.setenv(k, v)

    with pytest.raises(type(err)) as e:
        Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')

    assert str(err) == str(e.value)


def test_driver_api_client_ok(monkeypatch, requests_mock):
    requests_mock.post('https://api-host/api/v1/auth/login',
                       text='{"token": "token", "expiration": "2022-03-20T01:50:10.000Z"}')
    monkeypatch.setenv('CORELLIUM_API_HOST', 'api-host')
    monkeypatch.setenv('CORELLIUM_API_TOKEN', 'api-token')

    c = Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')

    assert c.api.req.headers['Authorization'] == 'Bearer api-token'


def test_driver_power_on_ok(monkeypatch):
    monkeypatch.setenv('CORELLIUM_API_HOST', 'api-host')
    monkeypatch.setenv('CORELLIUM_API_TOKEN', 'api-token')

    project = Project('1', 'Default Project')
    device = Device(name='rd1ae', type='automotive', flavor='kronos',
                    description='', model='kronos', peripherals=False, quotas={})
    instance = Instance(id='7f4f241c-821f-4219-905f-c3b50b0db5dd', state='on')
    root = Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')
    power = CorelliumPower(parent=root)

    with (patch.object(root.api, 'get_project', return_value=project),
          patch.object(root.api, 'get_device', return_value=device),
          patch.object(root.api, 'get_instance', side_effect=[None, instance]),
          patch.object(root.api, 'create_instance', return_value=instance)):
        power.on()


@pytest.mark.parametrize('mock_data', [
    ({'get_project': {'return_value': None}}),
    ({'get_instance': {'return_value': None}}),
    ({'create_instance': {'side_effect': CorelliumApiException('create error')}}),
])
def test_driver_power_on_error(monkeypatch, mock_data):
    monkeypatch.setenv('CORELLIUM_API_HOST', 'api-host')
    monkeypatch.setenv('CORELLIUM_API_TOKEN', 'api-token')

    project = Project('1', 'Default Project')
    instance = Instance(id='7f4f241c-821f-4219-905f-c3b50b0db5dd', state='on')
    root = Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')
    power = CorelliumPower(parent=root)

    with pytest.raises((CorelliumApiException, ValueError)):
        with (patch.object(root.api, 'get_project',
                           **mock_data.get('get_project', {'return_value': project})),
              patch.object(root.api, 'get_instance',
                           **mock_data.get('get_instance', {'return_value': instance})),
              patch.object(root.api, 'create_instance',
                           **mock_data.get('create_instance', {'return_value': instance}))):
            power.on()


def test_driver_power_off_ok(monkeypatch):
    monkeypatch.setenv('CORELLIUM_API_HOST', 'api-host')
    monkeypatch.setenv('CORELLIUM_API_TOKEN', 'api-token')

    project = Project('1', 'Default Project')
    instance = Instance(id='7f4f241c-821f-4219-905f-c3b50b0db5dd', state='on')
    root = Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')
    power = CorelliumPower(parent=root)

    with (patch.object(root.api, 'get_project', return_value=project),
          patch.object(root.api, 'set_instance_state', return_value=None),
          patch.object(root.api, 'get_instance',
                       side_effect=[instance, Instance(id=instance.id, state='off')])):
        power.off()


@pytest.mark.parametrize('mock_data',[
    ({'get_project': {'return_value': None}}),
    ({'get_instance': {'return_value': None}}),
    ({'destroy_instance': {'side_effect': CorelliumApiException('destroy error')}}),
])
def test_driver_power_off_error(monkeypatch, mock_data):
    monkeypatch.setenv('CORELLIUM_API_HOST', 'api-host')
    monkeypatch.setenv('CORELLIUM_API_TOKEN', 'api-token')

    project = Project('1', 'Default Project')
    instance = Instance(id='7f4f241c-821f-4219-905f-c3b50b0db5dd', state='on')
    root = Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')
    power = CorelliumPower(parent=root)

    with pytest.raises((CorelliumApiException, ValueError)):
        with (patch.object(root.api, 'get_project',
                           **mock_data.get('get_project', {'return_value': project})),
              patch.object(root.api, 'get_instance',
                           **mock_data.get('get_instance', {'side_effect': [instance, None]})),
              patch.object(root.api, 'destroy_instance',
                           **mock_data.get('destroy_instance', {'return_value': instance}))):
            power.off()


def test_driver_console_get_url_ok(monkeypatch):
    monkeypatch.setenv('CORELLIUM_API_HOST', 'api-host')
    monkeypatch.setenv('CORELLIUM_API_TOKEN', 'api-token')

    project = Project('1', 'Default Project')
    instance = Instance(id='7f4f241c-821f-4219-905f-c3b50b0db5dd', state='on')
    root = Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')
    console = CorelliumConsole(parent=root, url='')

    with (patch.object(root.api, 'get_project', return_value=project),
          patch.object(root.api, 'get_instance', return_value=instance),
          patch.object(root.api, 'get_instance_console_id', return_value='uart7-cons'),
          patch.object(root.api, 'get_instance_console_url', return_value='wss://mock')):
        assert 'wss://mock' == console.url


@pytest.mark.parametrize('mock_data',[
    ({'get_project': {'return_value': None}}),
    ({'get_instance': {'return_value': None}}),
    ({'get_instance_console_id': {'side_effect': ValueError('x')}}),
    ({'get_instance_console_url': {'side_effect': ValueError('x')}})
])
def test_driver_console_get_url_error(monkeypatch, mock_data):
    monkeypatch.setenv('CORELLIUM_API_HOST', 'api-host')
    monkeypatch.setenv('CORELLIUM_API_TOKEN', 'api-token')

    project = Project('1', 'Default Project')
    instance = Instance(id='7f4f241c-821f-4219-905f-c3b50b0db5dd', state='on')
    root = Corellium(project_id='1', device_name='jmp', device_flavor='kronos', device_os='1.0')
    console = CorelliumConsole(parent=root, url='')

    with pytest.raises((CorelliumApiException, ValueError)):
        with (patch.object(root.api, 'get_project',
                           **mock_data.get('get_project', {'return_value': project})),
              patch.object(root.api, 'get_instance',
                           **mock_data.get('get_instance', {'side_effect': [instance, None]})),
              patch.object(root.api, 'get_instance_console_id',
                           **mock_data.get('get_instance_console_id', {'return_value': 'uart7-cons'})),
              patch.object(root.api, 'get_instance_console_id',
                           **mock_data.get('get_instance_console_url', {'return_value': 'ws://mock'}))):
            assert console.url



from typing import Optional

import requests
import requests.exceptions

from .exceptions import CorelliumApiException
from .types import Device, Instance, Project


class ApiClient:
    """
    Corellium ReST API client used by the Corellium driver.
    """
    req: requests.Session

    def __init__(self, host: str, token: str) -> None:
        """
        Initializes a new client using the API token
        in all HTTP requests.
        """
        self.host = host
        self.token = token
        self.req = requests.Session()
        self.req.headers.update({'Authorization': f'Bearer {self.token}'})

    @property
    def baseurl(self) -> str:
        """
        Return the baseurl path for API calls.
        """
        return f'https://{self.host}/api'

    def get_project(self, project_ref: str = 'Default Project') -> Optional[Project]:
        """
        Retrieve a project based on project_ref, which is either its id or name.
        """
        data = None

        try:
            res = self.req.get(f'{self.baseurl}/v1/projects')
            data = res.json()
            res.raise_for_status()
        except requests.exceptions.RequestException as e:
            msgerr = data.get('error') if data is not None else str(e)

            raise CorelliumApiException(msgerr) from e

        for project in data:
            if project['name'] == project_ref or project['id'] == project_ref:
                return Project(id=project['id'], name=project['name'])

        return None

    def get_device(self, model: str) -> Optional[Device]:
        """
        Get a device spec from Corellium's list based on the model name.

        A device object is used to create a new virtual instance.
        """
        data = None

        try:
            res = self.req.get(f'{self.baseurl}/v1/models')
            data = res.json()
            res.raise_for_status()
        except requests.exceptions.RequestException as e:
            msgerr = data.get('error') if data is not None else str(e)

            raise CorelliumApiException(msgerr) from e

        for device in data:
            if device['model'] == model:
                return Device(**device) # ty: ignore[missing-argument]

        return None

    def create_instance(self, name: str, project: Project, device: Device, os_version: str, os_build: str) -> Instance:
        """
        Create a new virtual instance from a device spec.
        """
        data = None
        req_data = {
            'name': name,
            'project': project.id,
            'flavor': device.flavor,
            'os': os_version,
            'osbuild': os_build,
        }

        try:
            res = self.req.post(f'{self.baseurl}/v1/instances', json=req_data)
            data = res.json()
            res.raise_for_status()
        except requests.exceptions.RequestException as e:
            msgerr = data.get('error') if data is not None else str(e)

            raise CorelliumApiException(msgerr) from e

        return Instance(**data) # ty: ignore[missing-argument]

    def get_instance(self, instance_ref: str) -> Optional[Instance]:
        """
        Retrieve an existing instance by its name.

        Return None if it does not exist.
        """
        data = None

        try:
            res = self.req.get(f'{self.baseurl}/v1/instances')
            data = res.json()
            res.raise_for_status()
        except requests.exceptions.RequestException as e:
            msgerr = data.get('error') if data is not None else str(e)

            raise CorelliumApiException(msgerr) from e

        for instance in data:
            if instance['name'] == instance_ref or instance['id'] == instance_ref:
                return Instance(id=instance['id'], state=instance['state'])

        return None

    def set_instance_state(self, instance: Instance, instance_state: str) -> None:
        """
        Set the virtual instance state from corellium.

        Valid instance state values:

        - on
        - off
        - booting
        - deleting
        - creating
        - restoring
        - paused
        - rebooting
        - error
        """
        data = None
        req_data = {
            'state': instance_state
        }

        try:
            res = self.req.put(f'{self.baseurl}/v1/instances/{instance.id}/state', json=req_data)
            data = res.json() if res.status_code != 204 else None
            res.raise_for_status()
        except requests.exceptions.RequestException as e:
            msgerr = data.get('error') if data is not None else str(e)

            raise CorelliumApiException(msgerr) from e

    def destroy_instance(self, instance: Instance) -> None:
        """
        Delete a virtual instance.

        Does not return anything since Corellium's API return a HTTP 204 response.
        """
        try:
            res = self.req.delete(f'{self.baseurl}/v1/instances/{instance.id}')
            data = res.json() if res.status_code != 204 else None
            res.raise_for_status()
        except requests.exceptions.RequestException as e:
            msgerr = data.get('error') if data is not None else str(e)

            raise CorelliumApiException(msgerr) from e

    def get_instance_console_id(self, instance: Instance, console_name: str) -> Optional[str]:
        """
        Retrieve an instance's console id by its name.

        Return None if it does not exist.
        """
        data = None

        try:
            res = self.req.get(f'{self.baseurl}/v1/instances/{instance.id}')
            data = res.json()
            res.raise_for_status()
        except requests.exceptions.RequestException as e:
            msgerr = data.get('error') if data is not None else str(e)

            raise CorelliumApiException(msgerr) from e

        for console in data.get('consoles', []):
            if console['name'] == console_name:
                return console['id']

        return None

    def get_instance_console_url(self, instance: Instance, console_id: str) -> Optional[str]:
        """
        Get a a console URL (websocket) to stream logs from.
        """
        data = None

        try:
            res = self.req.get(f'{self.baseurl}/v1/instances/{instance.id}/console',
                               params={'type': console_id.replace('port-', '')})
            data = res.json()
            res.raise_for_status()
        except requests.exceptions.RequestException as e:
            msgerr = data.get('error') if data is not None else str(e)

            raise CorelliumApiException(msgerr) from e

        return data['url']



import os

import pytest

# import websockets
from .api import ApiClient
from .exceptions import CorelliumApiException
from .types import Device, Instance, Project


def fixture(path):
    """
    Load file contents from fixtures/$path.
    """
    cwd = os.path.dirname(os.path.abspath(__file__))
    fixtures_dir = f'{cwd}/../../fixtures'

    with open(f'{fixtures_dir}/{path}', 'r') as f:
        return f.read()


@pytest.mark.parametrize('project_name,data,has_results', [
    ('OtherProject', fixture('http/get-projects-200.json'), True),
    (None, fixture('http/get-projects-200.json'), True),
    ('notfound', fixture('http/get-projects-200.json'), False)
])
def test_get_project_ok(requests_mock, project_name, data, has_results):
    requests_mock.get('https://api-host/api/v1/projects', status_code=200, text=data)
    api = ApiClient('api-host', 'api-token')

    args = []
    if project_name:
        args.append(project_name)
    project = api.get_project(*args)

    if has_results:
        assert project is not None
        assert project.name == project_name if project_name is not None else 'Default Project'
    else:
        assert project is None


@pytest.mark.parametrize(
    'status_code,data,msg',
    [
        (403, fixture('http/403.json'), 'Invalid or missing authorization token'),
        (404, fixture('http/get-projects-404.json'), ''),
    ])
def test_get_project_error(requests_mock, status_code, data, msg):
    requests_mock.get('https://api-host/api/v1/projects', status_code=status_code, text=data)
    api = ApiClient('api-host', 'api-token')

    with pytest.raises(CorelliumApiException) as e:
        api.get_project()

    assert msg in str(e.value)


@pytest.mark.parametrize('model,data,has_results', [
    ('rpi4b', fixture('http/get-models-200.json'), True),
    ('notfound', fixture('http/get-models-200.json'), False)
])
def test_get_device_ok(requests_mock, model, data, has_results):
    requests_mock.get('https://api-host/api/v1/models', status_code=200, text=data)
    api = ApiClient('api-host', 'api-token')

    device = api.get_device(model)

    if has_results:
        assert device is not None
        assert device.model == model
    else:
        assert device is None


@pytest.mark.parametrize(
    'status_code,data,msg',
    [
        (403, fixture('http/403.json'), 'Invalid or missing authorization token'),
    ])
def test_get_device_error(requests_mock, status_code, data, msg):
    requests_mock.get('https://api-host/api/v1/models', status_code=status_code, text=data)
    api = ApiClient('api-host', 'api-token')

    with pytest.raises(CorelliumApiException) as e:
        api.get_device('mymodel')

    assert msg in str(e.value)


def test_create_instance_ok(requests_mock):
    data = fixture('http/create-instance-200.json')
    requests_mock.post('https://api-host/api/v1/instances', status_code=200, text=data)
    api = ApiClient('api-host', 'api-token')

    project = Project('d59db33d-27bd-4b22-878d-49e4758a648e', 'Default Project')
    device = Device(name='rd1ae', type='automotive', flavor='kronos',
                    description='', model='kronos', peripherals=False, quotas={})
    instance = api.create_instance('my-instance', project, device, '1.1.1', 'Critical Application Monitor (Baremetal)')

    assert instance is not None
    assert instance.id


@pytest.mark.parametrize(
    'status_code,data,msg',
    [
        (403, fixture('http/403.json'), 'Invalid or missing authorization token'),
        (400, fixture('http/create-instance-400.json'), 'Unsupported device model'),
    ])
def test_create_instance_error(requests_mock, status_code, data, msg):
    requests_mock.post('https://api-host/api/v1/instances', status_code=status_code, text=data)
    api = ApiClient('api-host', 'api-token')

    with pytest.raises(CorelliumApiException) as e:
        project = Project('d59db33d-27bd-4b22-878d-49e4758a648e', 'Default Project')
        device = Device(name='rd1ae', type='automotive', flavor='kronos',
                        description='', model='kronos', peripherals=False, quotas={})
        api.create_instance('my-instance', project, device, '1.1.1', 'Critical Application Monitor (Baremetal)')

    assert msg in str(e.value)


def test_destroy_instance_state_ok(requests_mock):
    instance = Instance(id='d59db33d-27bd-4b22-878d-49e4758a648e')

    requests_mock.delete(f'https://api-host/api/v1/instances/{instance.id}', status_code=204, text='')
    api = ApiClient('api-host', 'api-token')
    api.destroy_instance(instance)


@pytest.mark.parametrize(
    'status_code,data,msg',
    [
        (403, fixture('http/403.json'), 'Invalid or missing authorization token'),
        (404, fixture('http/get-instance-404.json'), 'No instance associated with this value'),
    ])
def test_destroy_instance_error(requests_mock, status_code, data, msg):
    instance = Instance(id='d59db33d-27bd-4b22-878d-49e4758a648e')

    requests_mock.delete(f'https://api-host/api/v1/instances/{instance.id}', status_code=status_code, text=data)
    api = ApiClient('api-host', 'api-token')

    with pytest.raises(CorelliumApiException) as e:
        api.destroy_instance(instance)

    assert msg in str(e.value)


@pytest.mark.parametrize(
    'console_name,console_id',
    [
        ('Console 1', 'port-1-cons',),
        ('Console 10', None,),
        ('Console 2', 'port-2-cons',),
    ])
def test_get_instance_console_id_ok(requests_mock, console_name, console_id):
    data = fixture('http/get-instance-200.json')
    instance = Instance(id='d59db33d-27bd-4b22-878d-49e4758a648e')
    requests_mock.get(f'https://api-host/api/v1/instances/{instance.id}', status_code=200, text=data)
    api = ApiClient('api-host', 'api-token')

    current = api.get_instance_console_id(instance, console_name)

    assert console_id == current


@pytest.mark.parametrize(
    'status_code,data,msg',
    [
        (403, fixture('http/403.json'), 'Invalid or missing authorization token'),
        (404, fixture('http/get-instance-404.json'), 'No instance associated with this value'),
    ])
def test_get_instance_console_id_error(requests_mock, status_code, data, msg):
    instance = Instance(id='d59db33d-27bd-4b22-878d-49e4758a648e')
    requests_mock.get(f'https://api-host/api/v1/instances/{instance.id}',
                      status_code=status_code, text=data)
    api = ApiClient('api-host', 'api-token')

    with pytest.raises(CorelliumApiException) as e:
        api.get_instance_console_id(instance, 'Console 1')

    assert msg in str(e.value)


def test_get_instance_console_url_ok(requests_mock):
    console_id = 'port-1-cons'
    data = fixture('http/get-instance-console-url-200.json')
    instance = Instance(id='d59db33d-27bd-4b22-878d-49e4758a648e')
    requests_mock.get(f'https://api-host/api/v1/instances/{instance.id}/console?type=1-cons',
                      status_code=200, text=data)
    api = ApiClient('api-host', 'api-token')

    current = api.get_instance_console_url(instance, console_id)
    expected = 'wss://api-host/port-cons-1'

    assert expected == current


@pytest.mark.parametrize(
    'status_code,data,msg',
    [
        (400, fixture('http/get-instance-console-url-400.json'), 'not a valid type'),
        (403, fixture('http/403.json'), 'Invalid or missing authorization token'),
        (404, fixture('http/get-instance-404.json'), 'No instance associated with this value'),
    ])
def test_get_instance_console_url_error(requests_mock, status_code, data, msg):
    console_id = 'port-1-cons'
    instance = Instance(id='d59db33d-27bd-4b22-878d-49e4758a648e')
    requests_mock.get(f'https://api-host/api/v1/instances/{instance.id}/console?type=1-cons',
                      status_code=status_code, text=data)
    api = ApiClient('api-host', 'api-token')

    with pytest.raises(CorelliumApiException) as e:
        api.get_instance_console_url(instance, console_id)

    assert msg in str(e.value)



"""
Corellium API client exceptions module
"""
from jumpstarter.common.exceptions import JumpstarterException


class CorelliumApiException(JumpstarterException):
    """
    Exception raised when something goes wrong with Corellium's API.
    """



"""
Jumpstarter corellium driver(s) implementation module.
"""

import os
import time
from collections.abc import AsyncGenerator
from dataclasses import dataclass, field
from typing import Dict, Optional

from jumpstarter_driver_network.driver import WebsocketNetwork
from jumpstarter_driver_power.driver import PowerReading, VirtualPowerInterface

from .corellium.api import ApiClient
from .corellium.types import Instance
from jumpstarter.common import exceptions as jmp_exceptions
from jumpstarter.driver import Driver, export


@dataclass(kw_only=True)
class Corellium(Driver):
    """
    Corellium top-level driver.
    """

    _api: ApiClient = field(init=False)
    project_id: str
    device_name: str
    device_flavor: str
    device_os: str = field(default="1.1.1")
    device_build: str = field(default="Critical Application Monitor (Baremetal)")
    console_name: str = field(default="Primary Compute Non-Secure")

    @classmethod
    def client(cls) -> str:
        """
        Return the driver's client.
        """
        return "jumpstarter_driver_corellium.client.CorelliumClient"

    def __post_init__(self) -> None:
        """
        Post initialization method.

        It will check for the following environment variables:

        - CORELLIUM_API_HOST
        - CORELLIUM_API_TOKEN

        Raises an exception in case these variables are not set or empty.

        Additionally, it also sets up some internal objects/varibales such as:

        - Corellium API client instance
        - Children jumpstarter drives
        """
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        api_host = self.get_env_var("CORELLIUM_API_HOST")
        api_token = self.get_env_var("CORELLIUM_API_TOKEN")
        self._api = ApiClient(api_host, api_token)

        self.children["power"] = CorelliumPower(parent=self)
        self.children["serial"] = CorelliumConsole(parent=self, url="")

    def get_env_var(self, name: str) -> str:
        """
        Return an env var and raise an exception if said
        var does not exist or is empty.
        """
        value = os.environ.get(name)

        if value is None:
            raise jmp_exceptions.ConfigurationError(f'Missing "{name}" environment variable')

        value = value.strip()

        if len(value) == 0:
            raise jmp_exceptions.ConfigurationError(f'"{name}" environment variable is empty')

        return value

    @property
    def api(self) -> ApiClient:
        """
        Return the Corellium API client object.
        """
        return self._api


@dataclass(kw_only=True)
class CorelliumPower(VirtualPowerInterface, Driver):
    """
    Power driver implementation for corellium virtual devices.

    This driver will create and destroy virtual instances.
    """

    parent: Corellium

    def get_timeout_opts(self) -> Dict[str, int]:
        """
        Return config/opts to be used when waiting for Corellium's API.
        """
        return {
            "retries": int(os.environ.get("CORELLIUM_API_RETRIES", 12)),
            "interval": os.environ.get("CORELLIUM_API_INTERVAL", 5),
        }

    def wait_instance(self, current: Instance, desired: Optional[Instance]):
        """
        Wait for `current` instance to reach the same state as the `desired` instance.

        Desired can also be set to None, which means the instance should not exist.
        """
        opts = self.get_timeout_opts()
        counter = 0

        while True:
            if counter >= opts["retries"]:
                raise ValueError(f"Instance took too long to be reach the desired state: {current}")

            if self.parent.api.get_instance(current.id) == desired:
                break

            counter += 1
            time.sleep(opts["interval"])

    @export
    def on(self) -> None:
        """
        Power a Corellium virtual device on.

        It will create an instance if one does not exist, it will just power the existing one on otherwise.
        """
        self.logger.info("Corellium Device:")
        self.logger.info(f"\tDevice Name: {self.parent.device_name}")
        self.logger.info(f"\tDevice Flavor: {self.parent.device_flavor}")
        self.logger.info(f"\tDevice OS Version: {self.parent.device_os}")

        project = self.parent.api.get_project(self.parent.project_id)
        if project is None:
            raise ValueError(f"Unable to fetch project: {self.parent.project_id}")
        self.logger.info(f"Using project: {project.name}")

        device = self.parent.api.get_device(self.parent.device_flavor)
        if device is None:
            raise ValueError("Unable to find a device for this model: {self.parent.device_model}")
        self.logger.info(f"Using device spec: {device.name}")

        # retrieve an existing instance first
        instance = self.parent.api.get_instance(self.parent.device_name)
        if instance:
            self.parent.api.set_instance_state(instance, "on")
        # create a new one otherwise
        else:
            opts = {}
            if self.parent.device_os:
                opts["os_version"] = self.parent.device_os
            if self.parent.device_build:
                opts["os_build"] = self.parent.device_build
            instance = self.parent.api.create_instance(self.parent.device_name, project, device, **opts)
        self.logger.info(f"Instance: {self.parent.device_name} (ID: {instance.id})")

        self.wait_instance(instance, Instance(id=instance.id, state="on"))

    @export
    def off(self, destroy: bool = False) -> None:
        """
        Destroy a Corellium virtual device/instance.
        """
        # fail if project does not exist
        project = self.parent.api.get_project(self.parent.project_id)
        if project is None:
            raise ValueError(f"Unable to fetch project: {self.parent.project_id}")

        # get instance and fail if instance does not exist
        instance = self.parent.api.get_instance(self.parent.device_name)
        if instance is None:
            raise ValueError("Instance does not exist")

        self.parent.api.set_instance_state(instance, "off")
        self.wait_instance(instance, Instance(id=instance.id, state="off"))

        if destroy:
            self.parent.api.destroy_instance(instance)
            self.wait_instance(instance, None)

    @export
    def read(self) -> AsyncGenerator[PowerReading, None]:
        raise NotImplementedError


@dataclass(kw_only=True)
class CorelliumConsole(WebsocketNetwork):
    """
    A serial console driver that uses a network websocket to connect
    to the remote virtual instance console.
    """

    parent: Corellium
    baudrate: int = field(default=115200)

    @classmethod
    def client(cls) -> str:
        """
        Use pyserial client to re-use its console implementation.
        """
        return "jumpstarter_driver_pyserial.client.PySerialClient"

    @property
    def url(self) -> str:
        """
        Retrieve console url from Corellium's API to be used by
        other drivers, such as the PySerial one.

        Overwrites the parent's url property so the required
        API calls are only invoked when the property is used.
        """
        project = self.parent.api.get_project(self.parent.project_id)
        if project is None:
            raise ValueError(f"Unable to fetch project: {self.parent.project_id}")

        # get instance and fail if instance does not exist
        instance = self.parent.api.get_instance(self.parent.device_name)
        if instance is None:
            raise ValueError("Instance does not exist or is powered off")

        console_id = self.parent.api.get_instance_console_id(instance, self.parent.console_name)
        if console_id is None:
            raise ValueError('Console ID not found for "{self.patent.console_name}"')
        console_url = self.parent.api.get_instance_console_url(instance, console_id)

        return console_url

    @url.setter
    def url(self, value: str) -> None:
        pass



"""
Corellium API types.
"""
from dataclasses import dataclass, field
from typing import Optional


@dataclass
class Project:
    """
    Dataclass that represents a Corellium project.
    """
    id: str
    name: str


@dataclass
class Device:
    """
    Dataclass to represent a Corellium Device.

    A device object is used to create virtual instances.
    """
    name: str
    type: str
    flavor: str
    description: str
    model: str
    peripherals: bool
    quotas: dict


@dataclass
class Instance:
    """
    Virtual instance dataclass.
    """
    id: str
    state: Optional[str] = field(default=None)
